#!/usr/local/bin/ruby
# Copyright (c) 2001 Brian Fundakowski Feldman
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# $Id: bonklenet.rb,v 1.1 2001/02/28 05:11:58 green Exp $
require 'bonkle'
require 'socket'
require 'thread'

module Bonkle
	module Message
		def Message.read(socket)
			m = socket.sysread(2)
			if m.size != 2
				return nil
			end
			socket.sysread(m.unpack('n')[0])
		end
		def Message.write(socket, message)
			socket.syswrite([message.size].pack('n').concat(message))
		end
		SUCCESS = 0
		FAILURE = 1
		@@ecodes = {
			0 => ' ',		# success
			1 => ' error: '	# failure
		}
		def Message.swrite(socket, ecode, message)
			message = ecode.to_s + @@ecodes[ecode] + message
			socket.syswrite([message.size].pack('n').concat(message))
		end
	end
	class Server
		class User
			[:name, :words, :status, :commandq].each {|att| attr_reader att}
			def initialize(sock, game)
				@socket = sock
				@game = game
				@name = ''
				@words = []
				@status = :init
				@starttime = Time.now
				@commandq = Queue.new
				@stop = false
				@thread = Thread.new {
					begin
						Message.write(@socket,
						  "info Hello, please introduce yourself.")
						until @stop
							process_event
						end
					rescue Exception => e
						$stderr.puts(e)
					ensure
						Thread.exit
					end
				}
			end
			def select_wait
				select([@socket], nil, nil, 1)
			end
			private :select_wait
			def process_event
				if !@commandq.empty?
					command = @commandq.shift
					case command[0]
					when :goodbye
						begin
							Message.write(@socket, "goodbye " + command[1])
						ensure
							@socket.close
							@stop = true
							if @status != :init
								@game.remove_user(self)
							end
							return
						end
					when :reset
						@words.empty
						Message.swrite(@socket, Message::SUCCESS,
						  "new game started")
					end
					return
				end

				if !select_wait
					return
				end
				case @status
				when :init
					@message = Message.read(@socket)
					if @message =~ /^name\s*(.+)$/
						@name = $1
						if @name !~ /^[[:print:]]*$/ || @name.length > 20
							@name = ''
						end
					end
					if @name.empty?
						Message.swrite(@socket, Message::FAILURE,
						  "need valid name")
						return
					elsif !@game.add_user(self)
						Message.swrite(@socket, Message::FAILURE,
						  "name already in use")
						return
					end
					@status = :run
					Message.swrite(@socket, Message::SUCCESS,
					  "name set")
				when :run
					@message = Message.read(@socket)
					case @message
					when /^timeleft(\s|$)/
						Message.swrite(@socket, Message::SUCCESS,
						  format("%.4f", @game.endtime - Time.now))
					when /^fboard(\s|$)/
						Message.swrite(@socket, Message::SUCCESS,
						  "pre-formatted Bonkle board:\n" +
						  @game.board.to_s)
					when /^boardsize(\s|$)/
						Message.swrite(@socket, Message::SUCCESS,
						  @game.board.width.to_s)
					when /^board(\s|$)/
						Message.swrite(@socket, Message::SUCCESS,
						  @game.board.dice_array.join(' '))
					when /^words(\s|$)/
						Message.swrite(@socket, Message::SUCCESS,
						  @words.join(' '))
					when /^word\s+(.+)$/
						w = $1.chomp.downcase
						if !@game.board.has_word?(w)
							Message.swrite(@socket, Message::FAILURE,
							  'word not found in board')
						elsif @words.include?(w)
							Message.swrite(@socket, Message::FAILURE,
							  'word already recorded')
						elsif !@game.dict.has_word?(w)
							Message.swrite(@socket, Message::FAILURE,
							  'word not recognized')
						else
							@words.push(w)
							Message.swrite(@socket, Message::SUCCESS,
							  'word now recorded')
						end
					else
						Message.swrite(@socket, Message::FAILURE,
						  'invalid command')
					end
				end
			end
			def queue_event(*event)
				@commandq.push(event)
			end
			def wait
				@thread.join
			end
			def goodbye(message)
				queue_event(:goodbye, message)
			end
		end
		# uses a Hash for options
		def initialize(opts = {})
			@game = Game.new
			@stop = false
			opts[:host] ||= '0'
			opts[:port] ||= 31782
			@socket = TCPServer.new(opts[:host], opts[:port])
			@thread = Thread.new {
				begin
					until @stop
						process_events
					end
					@game.each_user {|user|
						user.goodbye('BonkleServer shut down')
					}
					@game.each_user {|user|
						user.wait
					}
					@socket.close
				rescue Exception => e
					$stderr.puts(e)
				ensure
					Thread.exit
				end
			}
		end
		def stop
			@stop = true
			@thread.join
		end
		def process_events
			if select([@socket], [], [], 1)
				begin
					s = @socket.accept
				rescue
					return
				end
				User.new(s, @game)
			end
		end
		private :process_events
	end
	class Game
		attr_reader :dict
		attr_reader :endtime
		attr_reader :mutex
		attr_accessor :gamelength
		def initialize
			@board = Bonkle::Board.new
			@dict = Bonkle::Dict.new
			@gamelength = 120
			@users = {}
			@mutex = Mutex.new
			@starttime = Time.now
			@endtime = @starttime + @gamelength
		end
		def reset
			@starttime = Time.now
			@endtime = @starttime + @gamelength
			@board.reset
		end
		def add_user(user)
			@mutex.synchronize {
				if @users.has_key?(user.name)
					false
				else
					@users[user.name] = user
				end
			}
		end
		def remove_user(user)
			@mutex.synchronize {
				@users.delete(user.name)
			}
		end
		def each_user
			@mutex.synchronize {@users.dup}.each {|key, value| yield value}
		end
		def has_word?(word)
			@board.has_word?(word)
		end
		def finished?
			Time.now >= @endtime
		end
		def board
			@board
		end
	end
end
